/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import { BST, BSTNode } from './bst';
import type { AVLTreeNodeNested, AVLTreeOptions, BiTreeDeleteResult, BTNKey } from '../../types';
import { BTNCallback } from '../../types';
import { IBinaryTree } from '../../interfaces';
export declare class AVLTreeNode<V = any, N extends AVLTreeNode<V, N> = AVLTreeNodeNested<V>> extends BSTNode<V, N> {
    height: number;
    constructor(key: BTNKey, value?: V);
}
export declare class AVLTree<V = any, N extends AVLTreeNode<V, N> = AVLTreeNode<V, AVLTreeNodeNested<V>>> extends BST<V, N> implements IBinaryTree<V, N> {
    /**
     * This is a constructor function for an AVL tree data structure in TypeScript.
     * @param {AVLTreeOptions} [options] - The `options` parameter is an optional object that can be passed to the
     * constructor of the AVLTree class. It allows you to customize the behavior of the AVL tree by providing different
     * options.
     */
    constructor(options?: AVLTreeOptions);
    /**
     * The function creates a new AVL tree node with the specified key and value.
     * @param {BTNKey} key - The key parameter is the key value that will be associated with
     * the new node. It is used to determine the position of the node in the binary search tree.
     * @param [value] - The parameter `value` is an optional value that can be assigned to the node. It is of
     * type `V`, which means it can be any value that is assignable to the `value` property of the
     * node type `N`.
     * @returns a new AVLTreeNode object with the specified key and value.
     */
    createNode(key: BTNKey, value?: V): N;
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The add method of the superclass (BST) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     */
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The add method of the superclass (BST) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     *
     * The function overrides the add method of a class, adds a key-value pair to a data structure, and
     * balances the structure if necessary.
     * @param {BTNKey | N | null | undefined} keyOrNode - The `keyOrNode` parameter can be of type
     * `BTNKey`, `N`, `null`, or `undefined`.
     * @param {V} [value] - The `value` parameter is the value associated with the key that is being
     * added to the binary search tree.
     * @returns The method is returning either a node (N) or undefined.
     */
    add(keyOrNode: BTNKey | N | null | undefined, value?: V): N | undefined;
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The delete method of the superclass (BST) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     */
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The delete method of the superclass (BST) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     *
     * The function overrides the delete method of a binary tree, performs the deletion, and then
     * balances the tree if necessary.
     * @param identifier - The `identifier` parameter is the value or condition used to identify the
     * node(s) to be deleted from the binary tree. It can be of any type and is the return type of the
     * `callback` function.
     * @param {C} callback - The `callback` parameter is a function that will be called for each node
     * that is deleted from the binary tree. It is an optional parameter and if not provided, it will
     * default to the `_defaultOneParamCallback` function. The `callback` function should have a single
     * parameter of type `N
     * @returns The method is returning an array of `BiTreeDeleteResult<N>`.
     */
    delete<C extends BTNCallback<N>>(identifier: ReturnType<C>, callback?: C): BiTreeDeleteResult<N>[];
    /**
     * The `_swap` function swaps the key, value, and height properties between two nodes in a binary
     * tree.
     * @param {BTNKey | N | undefined} srcNode - The `srcNode` parameter represents the source node that
     * needs to be swapped with the destination node. It can be of type `BTNKey`, `N`, or `undefined`.
     * @param {BTNKey | N  | undefined} destNode - The `destNode` parameter represents the destination
     * node where the values from the source node will be swapped to.
     * @returns either the `destNode` object if both `srcNode` and `destNode` are defined, or `undefined`
     * if either `srcNode` or `destNode` is undefined.
     */
    protected _swap(srcNode: BTNKey | N | undefined, destNode: BTNKey | N | undefined): N | undefined;
    /**
     * Time Complexity: O(1) - constant time, as it performs a fixed number of operations.
     * Space Complexity: O(1) - constant space, as it only uses a constant amount of memory.
     */
    /**
     * Time Complexity: O(1) - constant time, as it performs a fixed number of operations.
     * Space Complexity: O(1) - constant space, as it only uses a constant amount of memory.
     *
     * The function calculates the balance factor of a node in a binary tree.
     * @param {N} node - The parameter "node" represents a node in a binary tree data structure.
     * @returns the balance factor of a given node. The balance factor is calculated by subtracting the
     * height of the left subtree from the height of the right subtree.
     */
    protected _balanceFactor(node: N): number;
    /**
     * Time Complexity: O(1) - constant time, as it performs a fixed number of operations.
     * Space Complexity: O(1) - constant space, as it only uses a constant amount of memory.
     */
    /**
     * Time Complexity: O(1) - constant time, as it performs a fixed number of operations.
     * Space Complexity: O(1) - constant space, as it only uses a constant amount of memory.
     *
     * The function updates the height of a node in a binary tree based on the heights of its left and
     * right children.
     * @param {N} node - The parameter "node" represents a node in a binary tree data structure.
     */
    protected _updateHeight(node: N): void;
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The method traverses the path from the inserted node to the root.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     */
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The method traverses the path from the inserted node to the root.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     *
     * The `_balancePath` function is used to update the heights of nodes and perform rotation operations
     * to restore balance in an AVL tree after inserting a node.
     * @param {N} node - The `node` parameter in the `_balancePath` function represents the node in the
     * AVL tree that needs to be balanced.
     */
    protected _balancePath(node: N): void;
    /**
     * Time Complexity: O(1) - constant time, as these methods perform a fixed number of operations.
     * Space Complexity: O(1) - constant space, as they only use a constant amount of memory.
     */
    /**
     * Time Complexity: O(1) - constant time, as these methods perform a fixed number of operations.
     * Space Complexity: O(1) - constant space, as they only use a constant amount of memory.
     *
     * The function `_balanceLL` performs a left-left rotation to balance a binary tree.
     * @param {N} A - A is a node in a binary tree.
     */
    protected _balanceLL(A: N): void;
    /**
     * Time Complexity: O(1) - constant time, as these methods perform a fixed number of operations.
     * Space Complexity: O(1) - constant space, as they only use a constant amount of memory.
     */
    /**
     * Time Complexity: O(1) - constant time, as these methods perform a fixed number of operations.
     * Space Complexity: O(1) - constant space, as they only use a constant amount of memory.
     *
     * The `_balanceLR` function performs a left-right rotation to balance a binary tree.
     * @param {N} A - A is a node in a binary tree.
     */
    protected _balanceLR(A: N): void;
    /**
     * Time Complexity: O(1) - constant time, as these methods perform a fixed number of operations.
     * Space Complexity: O(1) - constant space, as they only use a constant amount of memory.
     */
    /**
     * Time Complexity: O(1) - constant time, as these methods perform a fixed number of operations.
     * Space Complexity: O(1) - constant space, as they only use a constant amount of memory.
     *
     * The function `_balanceRR` performs a right-right rotation to balance a binary tree.
     * @param {N} A - A is a node in a binary tree.
     */
    protected _balanceRR(A: N): void;
    /**
     * Time Complexity: O(1) - constant time, as these methods perform a fixed number of operations.
     * Space Complexity: O(1) - constant space, as they only use a constant amount of memory.
     */
    /**
     * Time Complexity: O(1) - constant time, as these methods perform a fixed number of operations.
     * Space Complexity: O(1) - constant space, as they only use a constant amount of memory.
     *
     * The function `_balanceRL` performs a right-left rotation to balance a binary tree.
     * @param {N} A - A is a node in a binary tree.
     */
    protected _balanceRL(A: N): void;
}
