/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import type { BSTComparator, BSTNodeNested, BSTOptions, BTNCallback, BTNKey } from '../../types';
import { CP, IterationType } from '../../types';
import { BinaryTree, BinaryTreeNode } from './binary-tree';
import { IBinaryTree } from '../../interfaces';
export declare class BSTNode<V = any, N extends BSTNode<V, N> = BSTNodeNested<V>> extends BinaryTreeNode<V, N> {
    parent?: N;
    constructor(key: BTNKey, value?: V);
    protected _left?: N;
    /**
     * Get the left child node.
     */
    get left(): N | undefined;
    /**
     * Set the left child node.
     * @param {N | undefined} v - The left child node.
     */
    set left(v: N | undefined);
    protected _right?: N;
    /**
     * Get the right child node.
     */
    get right(): N | undefined;
    /**
     * Set the right child node.
     * @param {N | undefined} v - The right child node.
     */
    set right(v: N | undefined);
}
export declare class BST<V = any, N extends BSTNode<V, N> = BSTNode<V, BSTNodeNested<V>>> extends BinaryTree<V, N> implements IBinaryTree<V, N> {
    /**
     * The constructor function initializes a binary search tree with an optional comparator function.
     * @param {BSTOptions} [options] - An optional object that contains additional configuration options
     * for the binary search tree.
     */
    constructor(options?: BSTOptions);
    protected _root?: N;
    /**
     * Get the root node of the binary tree.
     */
    get root(): N | undefined;
    /**
     * The function creates a new binary search tree node with the given key and value.
     * @param {BTNKey} key - The key parameter is the key value that will be associated with
     * the new node. It is used to determine the position of the node in the binary search tree.
     * @param [value] - The parameter `value` is an optional value that can be assigned to the node. It
     * represents the value associated with the node in a binary search tree.
     * @returns a new instance of the BSTNode class with the specified key and value.
     */
    createNode(key: BTNKey, value?: V): N;
    /**
     * Time Complexity: O(log n) - Average case for a balanced tree. In the worst case (unbalanced tree), it can be O(n).
     * Space Complexity: O(1) - Constant space is used.
     */
    /**
     * Time Complexity: O(log n) - Average case for a balanced tree. In the worst case (unbalanced tree), it can be O(n).
     * Space Complexity: O(1) - Constant space is used.
     *
     * The `add` function adds a new node to a binary search tree based on the provided key and value.
     * @param {BTNKey | N | null | undefined} keyOrNode - The `keyOrNode` parameter can be one of the
     * following types:
     * @param {V} [value] - The `value` parameter is an optional value that can be associated with the
     * key or node being added to the binary search tree.
     * @returns The method `add` returns a node (`N`) that was inserted into the binary search tree. If
     * no node was inserted, it returns `undefined`.
     */
    add(keyOrNode: BTNKey | N | null | undefined, value?: V): N | undefined;
    /**
     * Time Complexity: O(n log n) - Adding each element individually in a balanced tree.
     * Space Complexity: O(n) - Additional space is required for the sorted array.
     */
    /**
     * Time Complexity: O(n log n) - Adding each element individually in a balanced tree.
     * Space Complexity: O(n) - Additional space is required for the sorted array.
     *
     * The `addMany` function is used to efficiently add multiple keys or nodes with corresponding data
     * to a binary search tree.
     * @param {(BTNKey | N | undefined)[]} keysOrNodes - An array of keys or nodes to be added to the
     * binary search tree. Each element can be of type `BTNKey` (binary tree node key), `N` (binary tree
     * node), or `undefined`.
     * @param {(V | undefined)[]} [data] - An optional array of values to associate with the keys or
     * nodes being added. If provided, the length of the `data` array must be the same as the length of
     * the `keysOrNodes` array.
     * @param [isBalanceAdd=true] - A boolean flag indicating whether the tree should be balanced after
     * adding the nodes. The default value is `true`.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to use when adding multiple keys or nodes to the binary search tree. It has a
     * default value of `this.iterationType`, which means it will use the iteration type specified in the
     * current instance of the binary search tree
     * @returns The function `addMany` returns an array of nodes (`N`) or `undefined` values.
     */
    addMany(keysOrNodes: (BTNKey | N | undefined)[], data?: (V | undefined)[], isBalanceAdd?: boolean, iterationType?: IterationType): (N | undefined)[];
    /**
     * Time Complexity: O(log n) - Average case for a balanced tree.
     * Space Complexity: O(1) - Constant space is used.
     */
    /**
     * Time Complexity: O(log n) - Average case for a balanced tree.
     * Space Complexity: O(1) - Constant space is used.
     *
     * The `lastKey` function returns the key of the rightmost node in a binary tree, or the key of the
     * leftmost node if the comparison result is greater than.
     * @param {BTNKey | N | undefined} beginRoot - The `beginRoot` parameter is optional and can be of
     * type `BTNKey`, `N`, or `undefined`. It represents the starting point for finding the last key in
     * the binary tree. If not provided, it defaults to the root of the binary tree (`this.root`).
     * @param iterationType - The `iterationType` parameter is used to specify the type of iteration to
     * be performed. It can have one of the following values:
     * @returns the key of the rightmost node in the binary tree if the comparison result is less than,
     * the key of the leftmost node if the comparison result is greater than, and the key of the
     * rightmost node otherwise. If no node is found, it returns 0.
     */
    lastKey(beginRoot?: BTNKey | N | undefined, iterationType?: IterationType): BTNKey;
    /**
     * Time Complexity: O(log n) - Average case for a balanced tree.
     * Space Complexity: O(log n) - Space for the recursive call stack in the worst case.
     */
    /**
     * Time Complexity: O(log n) - Average case for a balanced tree.
     * Space Complexity: O(log n) - Space for the recursive call stack in the worst case.
     *
     * The function `getNodeByKey` searches for a node in a binary tree based on a given key, using
     * either recursive or iterative methods.
     * @param {BTNKey} key - The `key` parameter is the key value that we are searching for in the tree.
     * It is used to identify the node that we want to retrieve.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to use when searching for a node in the binary tree. It can have two possible
     * values:
     * @returns The function `getNodeByKey` returns a node (`N`) if a node with the specified key is
     * found in the binary tree. If no node is found, it returns `undefined`.
     */
    getNodeByKey(key: BTNKey, iterationType?: IterationType): N | undefined;
    /**
     * The function `ensureNotKey` returns the node corresponding to the given key if it is a node key,
     * otherwise it returns the key itself.
     * @param {BTNKey | N | undefined} key - The `key` parameter can be of type `BTNKey`, `N`, or
     * `undefined`.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to be performed. It has a default value of `IterationType.ITERATIVE`.
     * @returns either a node object (N) or undefined.
     */
    ensureNotKey(key: BTNKey | N | undefined, iterationType?: IterationType): N | undefined;
    /**
     * Time Complexity: O(log n) - Average case for a balanced tree. O(n) - Visiting each node once when identifier is not node's key.
     * Space Complexity: O(log n) - Space for the recursive call stack in the worst case.
     */
    /**
     * Time Complexity: O(log n) - Average case for a balanced tree. O(n) - Visiting each node once when identifier is not node's key.
     * Space Complexity: O(log n) - Space for the recursive call stack in the worst case.
     *
     * The function `getNodes` returns an array of nodes that match a given identifier, using either a
     * recursive or iterative approach.
     * @param {ReturnType<C> | undefined} identifier - The `identifier` parameter is the value that you
     * want to search for in the nodes of the binary tree. It can be of any type that is returned by the
     * callback function `C`.
     * @param {C} callback - The `callback` parameter is a function that takes a node of type `N` as its
     * argument and returns a value of type `ReturnType<C>`. The `C` type parameter represents a callback
     * function type that extends the `BTNCallback<N>` type. The `BTNCallback<N>` type is
     * @param [onlyOne=false] - A boolean flag indicating whether to stop searching after finding the
     * first node that matches the identifier. If set to true, the function will return an array
     * containing only the first matching node. If set to false (default), the function will continue
     * searching for all nodes that match the identifier and return an array containing
     * @param {BTNKey | N | undefined} beginRoot - The `beginRoot` parameter represents the starting node
     * for the traversal. It can be either a key value or a node object. If it is undefined, the
     * traversal will start from the root of the tree.
     * @param iterationType - The `iterationType` parameter determines the type of iteration to be
     * performed on the binary tree. It can have two possible values:
     * @returns The method returns an array of nodes (`N[]`).
     */
    getNodes<C extends BTNCallback<N>>(identifier: ReturnType<C> | undefined, callback?: C, onlyOne?: boolean, beginRoot?: BTNKey | N | undefined, iterationType?: IterationType): N[];
    /**
     * Time Complexity: O(log n) - Average case for a balanced tree. O(n) - Visiting each node once when identifier is not node's key.
     * Space Complexity: O(log n) - Space for the recursive call stack in the worst case.
     */
    /**
     * Time Complexity: O(log n) - Average case for a balanced tree. O(n) - Visiting each node once when identifier is not node's key.
     * Space Complexity: O(log n) - Space for the recursive call stack in the worst case.
     *
     * The `lesserOrGreaterTraverse` function traverses a binary tree and returns an array of nodes that
     * are either lesser or greater than a target node, depending on the specified comparison type.
     * @param {C} callback - The `callback` parameter is a function that will be called for each node
     * that satisfies the condition specified by the `lesserOrGreater` parameter. It takes a single
     * parameter of type `N` (the node type) and returns a value of any type.
     * @param {CP} lesserOrGreater - The `lesserOrGreater` parameter is used to determine whether to
     * traverse nodes that are lesser than, greater than, or equal to the `targetNode`. It is of type
     * `CP`, which is a custom type representing the comparison operator. The possible values for
     * `lesserOrGreater` are
     * @param {BTNKey | N | undefined} targetNode - The `targetNode` parameter represents the node in the
     * binary tree that you want to traverse from. It can be specified either by its key, by the node
     * object itself, or it can be left undefined to start the traversal from the root of the tree.
     * @param iterationType - The `iterationType` parameter determines the type of traversal to be
     * performed on the binary tree. It can have two possible values:
     * @returns The function `lesserOrGreaterTraverse` returns an array of values of type
     * `ReturnType<C>`, which is the return type of the callback function passed as an argument.
     */
    lesserOrGreaterTraverse<C extends BTNCallback<N>>(callback?: C, lesserOrGreater?: CP, targetNode?: BTNKey | N | undefined, iterationType?: IterationType): ReturnType<C>[];
    /**
     * Balancing Adjustment:
     * Perfectly Balanced Binary Tree: Since the balance of a perfectly balanced binary tree is already fixed, no additional balancing adjustment is needed. Any insertion or deletion operation will disrupt the perfect balance, often requiring a complete reconstruction of the tree.
     * AVL Tree: After insertion or deletion operations, an AVL tree performs rotation adjustments based on the balance factor of nodes to restore the tree's balance. These rotations can be left rotations, right rotations, left-right rotations, or right-left rotations, performed as needed.
     *
     * Use Cases and Efficiency:
     * Perfectly Balanced Binary Tree: Perfectly balanced binary trees are typically used in specific scenarios such as complete binary heaps in heap sort or certain types of Huffman trees. However, they are not suitable for dynamic operations requiring frequent insertions and deletions, as these operations often necessitate full tree reconstruction.
     * AVL Tree: AVL trees are well-suited for scenarios involving frequent searching, insertion, and deletion operations. Through rotation adjustments, AVL trees maintain their balance, ensuring average and worst-case time complexity of O(log n).
     */
    /**
     * Time Complexity: O(n) - Building a balanced tree from a sorted array.
     * Space Complexity: O(n) - Additional space is required for the sorted array.
     */
    /**
     * Time Complexity: O(n) - Building a balanced tree from a sorted array.
     * Space Complexity: O(n) - Additional space is required for the sorted array.
     *
     * The `perfectlyBalance` function balances a binary search tree by adding nodes in a way that
     * ensures the tree is perfectly balanced.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to use when building a balanced binary search tree. It can have two possible
     * values:
     * @returns The function `perfectlyBalance` returns a boolean value.
     */
    perfectlyBalance(iterationType?: IterationType): boolean;
    /**
     * Time Complexity: O(n) - Visiting each node once.
     * Space Complexity: O(log n) - Space for the recursive call stack in the worst case.
     */
    /**
     * Time Complexity: O(n) - Visiting each node once.
     * Space Complexity: O(log n) - Space for the recursive call stack in the worst case.
     *
     * The function checks if a binary tree is AVL balanced using either recursive or iterative approach.
     * @param iterationType - The `iterationType` parameter is used to determine the method of iteration
     * to check if the AVL tree is balanced. It can have two possible values:
     * @returns a boolean value.
     */
    isAVLBalanced(iterationType?: IterationType): boolean;
    protected _comparator: BSTComparator;
    protected _setRoot(v: N | undefined): void;
    /**
     * The function compares two values using a comparator function and returns whether the first value
     * is greater than, less than, or equal to the second value.
     * @param {BTNKey} a - The parameter "a" is of type BTNKey.
     * @param {BTNKey} b - The parameter "b" in the above code represents a BTNKey.
     * @returns a value of type CP (ComparisonResult). The possible return values are CP.gt (greater
     * than), CP.lt (less than), or CP.eq (equal).
     */
    protected _compare(a: BTNKey, b: BTNKey): CP;
}
