/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import type { BTNKey, TreeMultimapNodeNested, TreeMultimapOptions } from '../../types';
import { BiTreeDeleteResult, BTNCallback, IterationType } from '../../types';
import { IBinaryTree } from '../../interfaces';
import { AVLTree, AVLTreeNode } from './avl-tree';
export declare class TreeMultimapNode<V = any, N extends TreeMultimapNode<V, N> = TreeMultimapNodeNested<V>> extends AVLTreeNode<V, N> {
    count: number;
    /**
     * The constructor function initializes a BinaryTreeNode object with a key, value, and count.
     * @param {BTNKey} key - The `key` parameter is of type `BTNKey` and represents the unique identifier
     * of the binary tree node.
     * @param {V} [value] - The `value` parameter is an optional parameter of type `V`. It represents the value of the binary
     * tree node. If no value is provided, it will be `undefined`.
     * @param {number} [count=1] - The `count` parameter is a number that represents the number of times a particular value
     * occurs in a binary tree node. It has a default value of 1, which means that if no value is provided for the `count`
     * parameter when creating a new instance of the `BinaryTreeNode` class.
     */
    constructor(key: BTNKey, value?: V, count?: number);
}
/**
 * The only distinction between a TreeMultimap and a AVLTree lies in the ability of the former to store duplicate nodes through the utilization of counters.
 */
export declare class TreeMultimap<V = any, N extends TreeMultimapNode<V, N> = TreeMultimapNode<V, TreeMultimapNodeNested<V>>> extends AVLTree<V, N> implements IBinaryTree<V, N> {
    /**
     * The constructor function for a TreeMultimap class in TypeScript, which extends another class and sets an option to
     * merge duplicated values.
     * @param {TreeMultimapOptions} [options] - An optional object that contains additional configuration options for the
     * TreeMultimap.
     */
    constructor(options?: TreeMultimapOptions);
    private _count;
    get count(): number;
    /**
     * The function creates a new BSTNode with the given key, value, and count.
     * @param {BTNKey} key - The key parameter is the unique identifier for the binary tree node. It is used to
     * distinguish one node from another in the tree.
     * @param {N} value - The `value` parameter represents the value that will be stored in the binary search tree node.
     * @param {number} [count] - The "count" parameter is an optional parameter of type number. It represents the number of
     * occurrences of the value in the binary search tree node. If not provided, the count will default to 1.
     * @returns A new instance of the BSTNode class with the specified key, value, and count (if provided).
     */
    createNode(key: BTNKey, value?: V, count?: number): N;
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The add method of the superclass (AVLTree) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     */
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The add method of the superclass (AVLTree) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     *
     * The `add` function adds a new node to the tree multimap, updating the count if the key already
     * exists, and balances the tree if necessary.
     * @param {BTNKey | N | null | undefined} keyOrNode - The `keyOrNode` parameter can be one of the
     * following types:
     * @param {V} [value] - The `value` parameter represents the value associated with the key that is
     * being added to the tree. It is an optional parameter, so it can be omitted if not needed.
     * @param [count=1] - The `count` parameter is an optional parameter that specifies the number of
     * times the key-value pair should be added to the multimap. If not provided, the default value is 1.
     * @returns a node (`N`) or `undefined`.
     */
    add(keyOrNode: BTNKey | N | null | undefined, value?: V, count?: number): N | undefined;
    /**
     * Time Complexity: O(1) - constant time, as it performs basic pointer assignments.
     * Space Complexity: O(1) - constant space, as it only uses a constant amount of memory.
     */
    /**
     * Time Complexity: O(1) - constant time, as it performs basic pointer assignments.
     * Space Complexity: O(1) - constant space, as it only uses a constant amount of memory.
     *
     * The function adds a new node to a binary tree, either as the left child or the right child of a
     * given parent node.
     * @param {N | undefined} newNode - The `newNode` parameter represents the node that needs to be
     * added to the binary tree. It can be of type `N` (which represents a node in the binary tree) or
     * `undefined` if there is no node to add.
     * @param {BTNKey | N | undefined} parent - The `parent` parameter represents the parent node to
     * which the new node will be added as a child. It can be either a node object (`N`) or a key value
     * (`BTNKey`).
     * @returns The method `_addTo` returns either the `parent.left` or `parent.right` node that was
     * added, or `undefined` if no node was added.
     */
    protected _addTo(newNode: N | undefined, parent: BTNKey | N | undefined): N | undefined;
    /**
     * Time Complexity: O(k log n) - logarithmic time for each insertion, where "n" is the number of nodes in the tree, and "k" is the number of keys to be inserted. This is because the method iterates through the keys and calls the add method for each.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     */
    /**
     * Time Complexity: O(k log n) - logarithmic time for each insertion, where "n" is the number of nodes in the tree, and "k" is the number of keys to be inserted. This is because the method iterates through the keys and calls the add method for each.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     *
     * The function `addMany` takes an array of keys or nodes and adds them to the TreeMultimap,
     * returning an array of the inserted nodes.
     * @param {(BTNKey | N | undefined)[]} keysOrNodes - An array of keys or nodes. Each element can be
     * of type BTNKey, N, or undefined.
     * @param {V[]} [data] - The `data` parameter is an optional array of values that correspond to the
     * keys or nodes being added. It is used to associate data with each key or node being added to the
     * TreeMultimap. If provided, the length of the `data` array should be the same as the length of the
     * @returns The function `addMany` returns an array of nodes (`N`) or `undefined` values.
     */
    addMany(keysOrNodes: (BTNKey | N | undefined)[], data?: V[]): (N | undefined)[];
    /**
     * Time Complexity: O(n log n) - logarithmic time for each insertion, where "n" is the number of nodes in the tree. This is because the method calls the add method for each node.
     * Space Complexity: O(n) - linear space, as it creates an array to store the sorted nodes.
     */
    /**
     * Time Complexity: O(n log n) - logarithmic time for each insertion, where "n" is the number of nodes in the tree. This is because the method calls the add method for each node.
     * Space Complexity: O(n) - linear space, as it creates an array to store the sorted nodes.
     *
     * The `perfectlyBalance` function takes a sorted array of nodes and builds a balanced binary search
     * tree using either a recursive or iterative approach.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to use when building the balanced binary search tree. It can have two possible
     * values:
     * @returns a boolean value.
     */
    perfectlyBalance(iterationType?: IterationType): boolean;
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The delete method of the superclass (AVLTree) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     */
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The delete method of the superclass (AVLTree) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     *
     * The `delete` function in TypeScript is used to remove a node from a binary tree, taking into
     * account the count of the node and balancing the tree if necessary.
     * @param identifier - The identifier is the value or key that is used to identify the node that
     * needs to be deleted from the binary tree. It can be of any type that is returned by the callback
     * function.
     * @param {C} callback - The `callback` parameter is a function that is used to determine if a node
     * should be deleted. It is optional and defaults to a default callback function. The `callback`
     * function takes one parameter, which is the identifier of the node, and returns a value that is
     * used to identify the node to
     * @param [ignoreCount=false] - A boolean flag indicating whether to ignore the count of the node
     * being deleted. If set to true, the count of the node will not be considered and the node will be
     * deleted regardless of its count. If set to false (default), the count of the node will be
     * decremented by 1 and
     * @returns an array of `BiTreeDeleteResult<N>`.
     */
    delete<C extends BTNCallback<N>>(identifier: ReturnType<C>, callback?: C, ignoreCount?: boolean): BiTreeDeleteResult<N>[];
    /**
     * The clear() function clears the contents of a data structure and sets the count to zero.
     */
    clear(): void;
    /**
     * The `_swap` function swaps the key, value, count, and height properties between two nodes.
     * @param {BTNKey | N | undefined} srcNode - The `srcNode` parameter represents the source node from
     * which the values will be swapped. It can be of type `BTNKey`, `N`, or `undefined`.
     * @param {BTNKey | N | undefined} destNode - The `destNode` parameter represents the destination
     * node where the values from the source node will be swapped to.
     * @returns either the `destNode` object if both `srcNode` and `destNode` are defined, or `undefined`
     * if either `srcNode` or `destNode` is undefined.
     */
    protected _swap(srcNode: BTNKey | N | undefined, destNode: BTNKey | N | undefined): N | undefined;
}
