import { AbstractEdge, AbstractGraph, AbstractVertex } from './abstract-graph';
import type { VertexKey } from '../../types';
import { IGraph } from '../../interfaces';
export declare class UndirectedVertex<V = any> extends AbstractVertex<V> {
    /**
     * The constructor function initializes a vertex with an optional value.
     * @param {VertexKey} key - The `key` parameter is of type `VertexKey` and represents the identifier of the vertex. It is
     * used to uniquely identify the vertex within a graph or network.
     * @param {V} [value] - The "value" parameter is an optional parameter of type V. It is used to initialize the value of the
     * vertex. If no value is provided, the vertex will be initialized with a default value.
     */
    constructor(key: VertexKey, value?: V);
}
export declare class UndirectedEdge<E = number> extends AbstractEdge<E> {
    vertices: [VertexKey, VertexKey];
    /**
     * The constructor function creates an instance of a class with two vertex IDs, an optional weight, and an optional
     * value.
     * @param {VertexKey} v1 - The first vertex ID of the edge.
     * @param {VertexKey} v2 - The parameter `v2` is a `VertexKey`, which represents the identifier of the second vertex in a
     * graph edge.
     * @param {number} [weight] - The weight parameter is an optional number that represents the weight of the edge.
     * @param {E} [value] - The "value" parameter is an optional parameter of type E. It is used to store a value associated
     * with the edge.
     */
    constructor(v1: VertexKey, v2: VertexKey, weight?: number, value?: E);
}
export declare class UndirectedGraph<V = any, E = any, VO extends UndirectedVertex<V> = UndirectedVertex<V>, EO extends UndirectedEdge<E> = UndirectedEdge<E>> extends AbstractGraph<V, E, VO, EO> implements IGraph<V, E, VO, EO> {
    /**
     * The constructor initializes a new Map object to store edges.
     */
    constructor();
    protected _edges: Map<VO, EO[]>;
    get edges(): Map<VO, EO[]>;
    /**
     * The function creates a new vertex with an optional value and returns it.
     * @param {VertexKey} key - The `key` parameter is the unique identifier for the vertex. It is used to distinguish one
     * vertex from another in the graph.
     * @param [value] - The `value` parameter is an optional value that can be assigned to the vertex. If a value is provided,
     * it will be used as the value of the vertex. If no value is provided, the `key` parameter will be used as the value of
     * the vertex.
     * @returns The method is returning a new instance of the `UndirectedVertex` class, casted as type `VO`.
     */
    createVertex(key: VertexKey, value?: VO['value']): VO;
    /**
     * The function creates an undirected edge between two vertices with an optional weight and value.
     * @param {VertexKey} v1 - The parameter `v1` represents the first vertex of the edge.
     * @param {VertexKey} v2 - The parameter `v2` represents the second vertex of the edge.
     * @param {number} [weight] - The `weight` parameter is an optional number that represents the weight of the edge. If
     * no weight is provided, it defaults to 1.
     * @param [value] - The `value` parameter is an optional value that can be assigned to the edge. It can be of any type and
     * is used to store additional information or data associated with the edge.
     * @returns a new instance of the `UndirectedEdge` class, which is casted as type `EO`.
     */
    createEdge(v1: VertexKey, v2: VertexKey, weight?: number, value?: EO['value']): EO;
    /**
     * Time Complexity: O(|E|), where |E| is the number of edges incident to the given vertex.
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(|E|), where |E| is the number of edges incident to the given vertex.
     * Space Complexity: O(1)
     *
     * The function `getEdge` returns the first edge that connects two vertices, or null if no such edge exists.
     * @param {VO | VertexKey | null} v1 - The parameter `v1` represents a vertex or vertex ID. It can be of type `VO` (vertex
     * object), `null`, or `VertexKey` (a string or number representing the ID of a vertex).
     * @param {VO | VertexKey | null} v2 - The parameter `v2` represents a vertex or vertex ID. It can be of type `VO` (vertex
     * object), `null`, or `VertexKey` (vertex ID).
     * @returns an edge (EO) or null.
     */
    getEdge(v1: VO | VertexKey | null, v2: VO | VertexKey | null): EO | null;
    /**
     * Time Complexity: O(|E|), where |E| is the number of edges incident to the given vertex.
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(|E|), where |E| is the number of edges incident to the given vertex.
     * Space Complexity: O(1)
     *
     * The function removes an edge between two vertices in a graph and returns the removed edge.
     * @param {VO | VertexKey} v1 - The parameter `v1` represents either a vertex object (`VO`) or a vertex ID (`VertexKey`).
     * @param {VO | VertexKey} v2 - VO | VertexKey - This parameter can be either a vertex object (VO) or a vertex ID
     * (VertexKey). It represents the second vertex of the edge that needs to be removed.
     * @returns the removed edge (EO) if it exists, or null if either of the vertices (VO) does not exist.
     */
    deleteEdgeBetween(v1: VO | VertexKey, v2: VO | VertexKey): EO | null;
    /**
     * Time Complexity: O(|E|), where |E| is the number of edges incident to the given vertex.
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(|E|), where |E| is the number of edges incident to the given vertex.
     * Space Complexity: O(1)
     *
     * The deleteEdge function removes an edge between two vertices in a graph.
     * @param {EO} edge - The parameter "edge" is of type EO, which represents an edge in a graph.
     * @returns The method is returning either the removed edge (of type EO) or null if the edge was not found.
     */
    deleteEdge(edge: EO): EO | null;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function `degreeOf` returns the degree of a vertex in a graph, which is the number of edges connected to that
     * vertex.
     * @param {VertexKey | VO} vertexOrKey - The parameter `vertexOrKey` can be either a `VertexKey` or a `VO`.
     * @returns The function `degreeOf` returns the degree of a vertex in a graph. The degree of a vertex is the number of
     * edges connected to that vertex.
     */
    degreeOf(vertexOrKey: VertexKey | VO): number;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function returns the edges of a given vertex or vertex ID.
     * @param {VertexKey | VO} vertexOrKey - The parameter `vertexOrKey` can be either a `VertexKey` or a `VO`. A `VertexKey` is a
     * unique identifier for a vertex in a graph, while `VO` represents the type of the vertex.
     * @returns an array of edges.
     */
    edgesOf(vertexOrKey: VertexKey | VO): EO[];
    /**
     * Time Complexity: O(|V| + |E|), where |V| is the number of vertices and |E| is the number of edges.
     * Space Complexity: O(|E|)
     */
    /**
     * Time Complexity: O(|V| + |E|), where |V| is the number of vertices and |E| is the number of edges.
     * Space Complexity: O(|E|)
     *
     * The function "edgeSet" returns an array of unique edges from a set of edges.
     * @returns The method `edgeSet()` returns an array of type `EO[]`.
     */
    edgeSet(): EO[];
    /**
     * Time Complexity: O(|V| + |E|), where |V| is the number of vertices and |E| is the number of edges.
     * Space Complexity: O(|E|)
     */
    /**
     * Time Complexity: O(|V| + |E|), where |V| is the number of vertices and |E| is the number of edges.
     * Space Complexity: O(|E|)
     *
     * The function "getNeighbors" returns an array of neighboring vertices for a given vertex or vertex ID.
     * @param {VO | VertexKey} vertexOrKey - The parameter `vertexOrKey` can be either a vertex object (`VO`) or a vertex ID
     * (`VertexKey`).
     * @returns an array of vertices (VO[]).
     */
    getNeighbors(vertexOrKey: VO | VertexKey): VO[];
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function "getEndsOfEdge" returns the vertices at the ends of an edge if the edge exists in the graph, otherwise
     * it returns null.
     * @param {EO} edge - The parameter "edge" is of type EO, which represents an edge in a graph.
     * @returns The function `getEndsOfEdge` returns an array containing two vertices `[VO, VO]` if the edge exists in the
     * graph. If the edge does not exist, it returns `null`.
     */
    getEndsOfEdge(edge: EO): [VO, VO] | null;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     */
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function adds an edge to the graph by updating the adjacency list with the vertices of the edge.
     * @param {EO} edge - The parameter "edge" is of type EO, which represents an edge in a graph.
     * @returns a boolean value.
     */
    protected _addEdgeOnly(edge: EO): boolean;
}
